/* -*- C++ -*-
 * File: io.c
 * Copyright 2008 Alex Tutubalin <lexa@lexa.ru>
 * Created: Sat Mar  8 , 2008
 *
 * LibRaw internal I/O layer (implementation)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#define O_BINARY 0
#else
#include <io.h>
#endif
#include <fcntl.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#ifdef USE_MMAP_IO
#include <sys/mman.h>
#endif
#define IODEF_IMPL
#include "libraw/libraw_types.h"
#include "libraw/libraw_io.h"

#ifdef __cplusplus
extern "C" 
{
#endif

int datastream_pop(t_datastream *stream)
{
    if(stream->chain)
        {
            t_datastream *next = (t_datastream *)stream->chain;
            datastream_close(stream);
            stream->pos = next->pos;
            stream->size = next->size;
            stream->buf = next->buf;
            stream->mmap_size = next->mmap_size;
            stream->chain = next->chain;
            free(next);
        }
    return 0;
}

int datastream_pushbytes(void *data, size_t size, t_datastream *stream)
{
    // move current data to chain
    t_datastream *push = calloc(1,sizeof(t_datastream));
    if (!push) return ENOMEM;
    push->pos = stream->pos;
    push->size = stream->size;
    push->buf = stream->buf;
    push->mmap_size = stream->mmap_size;
    push->chain = stream->chain;
    stream->chain = push;

    // create buffer than move data;
    stream->buf = malloc(size);
    if(!stream->buf) return ENOMEM;
    memmove(stream->buf,data,size);
    stream->mmap_size = 0;
    stream->size = size;
    stream->pos = 0;
    return 0;
}

int datastream_pushfile(const char *fname, t_datastream *stream)
{
    // move current data to chain
    t_datastream *push ;
    t_datastream newstream;
    int ret;

    if(!(ret = datastream_fopen(fname,&newstream)))
        return ret;
    if ( !(push= calloc(1,sizeof(t_datastream))))
        {
            datastream_close(&newstream);
            return ENOMEM;
        }

    push->pos = stream->pos;
    push->size = stream->size;
    push->buf = stream->buf;
    push->mmap_size = stream->mmap_size;
    push->chain = stream->chain;
    stream->chain = push;

    // copy data
    
    stream->buf = newstream.buf;
    stream->size = newstream.size;
    stream->pos = newstream.pos;
    stream->mmap_size = newstream.mmap_size;

    return 0;
}



int datastream_fopen(const char *fname, t_datastream *stream)
{
    struct stat st;
    int file;
    assert(stream);
    stream->chain = 0;
    if(!fname) return ENOENT;
    if((file = open(fname,O_RDONLY|O_BINARY))<0)
        return errno;
    if(stat(fname,&st))
        return errno;
    if(st.st_size<1)
        return EINVAL;
#ifdef USE_MMAP_IO
    {
        int pgsz = getpagesize();
        stream->mmap_size = ((st.st_size + pgsz -1) / pgsz) * pgsz;
        stream->buf = mmap(NULL,stream->mmap_size,PROT_READ,MAP_PRIVATE,file,0);
        if(!stream->buf) return errno;
        stream->size = st.st_size;
    }
#else
    if(stream->buf) 
        free(stream->buf);
    if(!(stream->buf = (unsigned char*) malloc(st.st_size)))
        return ENOMEM;
    stream->pos = 0;
    if((stream->size = read(file,stream->buf,st.st_size))<0)
        return errno;
#endif
    close(file);
    return 0;
}

int datasteam_allocopen(size_t size, t_datastream *stream)
{
    assert(stream);
    if(stream->buf) 
        free(stream->buf);
    if(!(stream->buf = (unsigned char *) malloc(size)))
        return ENOMEM;
    stream->pos = 0;
    return 0;
}

void datastream_closeall(t_datastream *stream)
{
    if(stream->chain)
        {
            datastream_closeall(stream->chain);
            free(stream->chain);
        }
    datastream_close(stream);
}

void datastream_close(t_datastream *stream)
{
    assert(stream);
#ifdef USE_MMAP_IO
    if(stream->mmap_size && stream->buf)
        {
            munmap(stream->buf,stream->mmap_size);
            stream->mmap_size = 0;
            stream->buf = NULL;
        }
    else
#endif
    if(stream->buf) 
        free(stream->buf);
    stream->buf = NULL;
    stream->size = stream->pos = 0;
}

int datastream_getc(t_datastream *stream)
{
    assert(stream);
    if(stream->pos >= stream->size)
        return -1;
    return stream->buf[stream->pos++];
}

size_t datastream_fread(void *buf, size_t size, size_t n, t_datastream *stream)
{
    size_t to_read = size * n;
    assert(stream);
    if(to_read > stream->size - stream->pos)
        to_read = stream->size - stream->pos;
    if (to_read < 1)
        return 0 ;
    memmove(buf,stream->buf + stream->pos, to_read);
    stream->pos += to_read;
    return (to_read+size-1)/size;
}
int datastream_end(t_datastream *stream)
{
    assert(stream);
    return stream->pos >= stream->size;
}

off_t datastream_tell(t_datastream *stream)
{
    assert(stream);
    return stream->pos;
}

int datastream_seek(off_t pos, int whence, t_datastream *stream)
{
    assert(stream);
    switch (whence) 
        {
        case SEEK_SET:
            stream->pos = pos > stream->size ? stream->size : pos;
            return 0;
        case SEEK_CUR:
            if(pos < 0)
                {
                    if (pos < - stream->pos)
                        pos = - stream->pos;
                    stream->pos += pos;
                }
            else if (pos > 0)
                {
                    if(pos > stream->size - stream->pos)
                        pos = stream->size - stream->pos;
                    stream->pos += pos;
                }
            return 0;
        case SEEK_END:
            if(pos>0)
                stream->pos = stream->size;
            else if (-pos > stream->size)
                stream->pos = 0;
            else 
                stream->pos = stream->size + pos;
                    
            return 0;
        default:
            // ???
            return 0;
        }
}

int datastream_scanf(t_datastream *stream, const char * fmt, void *ptr)
{
    int scanf_res;
    assert(stream);
    if(stream->pos >= stream->size)
        return 0;
    scanf_res = sscanf((char*)(stream->buf+stream->pos),fmt,ptr);
    if(scanf_res>0)
        stream->pos++; // we really does not know actual count
    return scanf_res;
    
}

unsigned char * datastream_gets(unsigned char *buf, int sz,t_datastream *stream)
{
    unsigned char *psrc,*pdest;
    assert(stream);
    psrc = stream->buf+stream->pos;
    pdest = buf;
    while (((psrc - stream->buf)<stream->size)
           &&
           ((pdest - buf) < sz)
        )
        {
            *pdest = *psrc;
            if(*psrc == '\n')
                break;
            psrc++;
            pdest++;
        }
    if((psrc - stream->buf) < stream->size)
        psrc++;
    if( (pdest-buf) < sz)
        *(++pdest)=0;
    stream->pos = psrc - stream->buf;
    return buf;
}

#ifdef __cplusplus
};
#endif

